#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import re
from collections import deque
from typing import Set, Iterator, Iterable
from uuid import uuid4

import tenacity
import requests
import toposort
from libs.constants import RETRIES, TIMEOUT
from libs.context import Context
from libs.mixins import HeadersMixin
from libs.refresh_token import TokenRefresher
from libs.traverse_manifest import ManifestEntity
from osdu_api.libs.auth.authorization import authorize

logger = logging.getLogger()


class EntityNode(object):
    """
    This class represents entities and their links to parent and child ones.
    """
    __slots__ = ["srn", "system_srn", "entity_info", "children", "parents", "unprocessed"]

    SRN_REGEX = re.compile(
        r"(?<=\")surrogate-key:[\w\-\.\d]+(?=\")|(?<=\")[\w\-\.]+:[\w\-\.]+--[\w\-\.]+:[\w\-\.\d]+(?=\")")

    def __init__(self, srn, entity_info: ManifestEntity):
        self.srn = srn
        self.entity_info = entity_info
        self.system_srn = None
        self.children = set()
        self.parents = set()
        self.unprocessed = False

    def __repr__(self):
        return f"SRN: {self.srn}"

    @property
    def content(self) -> dict:
        return self.entity_info.entity

    @content.setter
    def content(self, value: dict):
        self.entity_info.entity = value

    def add_child(self, child_node: "EntityNode"):
        self.children.add(child_node)

    def add_parent(self, parent_node: "EntityNode"):
        self.parents.add(parent_node)

    def get_parent_srns(self) -> Set[str]:
        """
        Get list of parents' srns.
        """
        entity_content = json.dumps(self.content, separators=(",", ":"))
        parent_srns = set(self.SRN_REGEX.findall(entity_content))
        parent_srns.discard(self.srn)
        return parent_srns

    def replace_parents_surrogate_srns(self):
        """
        Replace surrogate parents' keys with system-generated ones in entity.
        """
        if not self.parents:
            return
        content = json.dumps(self.content)
        for parent in self.parents:
            if parent.system_srn:
                content = content.replace(parent.srn, parent.system_srn)
        self.content = json.loads(content)


class ManifestAnalyzer(HeadersMixin):
    """
    This class is for creating a queue for ingesting set of data, each piece of data can depend on
    another one, so we must prioritize the order of ingesting. The idea is to create a dependency
    graph and traverse it to get the right order of ingestion.

    The flow of prioritizing entities could be described as:
    1. Fill graph's nodes with entities (self._fill_srn_node_table())
    2. Create links between nodes (self._fill_nodes_edges())
    3. Mark unprocessed nodes if they are orphaned or dependant on orphaned nodes (self._find_unprocessed_nodes())
    4. Return prioritized queue for ingesting (self.entity_queue())
    """

    def __init__(
        self,
        entities: Iterable[ManifestEntity],
        storage_service_url: str,
        token_refresher: TokenRefresher,
        context: Context
    ):
        super().__init__(context)
        self.entities = entities
        self.storage_service_url = storage_service_url
        self.token_refresher = token_refresher
        self.srn_node_table = dict()
        self.processed_entities = []

        # used as a root for all orphan entities
        empty_entity_info = ManifestEntity("", {})
        self.unprocessed_entities_parent = EntityNode(srn=str(uuid4()),
                                                      entity_info=empty_entity_info)
        self.unprocessed_entities = set()

        self._fill_srn_node_table()
        self._fill_nodes_parents()
        self._find_unprocessed_nodes()

    def _create_entity_node(self, entity: ManifestEntity):
        srn = entity.entity.get("id", f"surrogate-key:{str(uuid4())}")
        self.srn_node_table[srn] = EntityNode(srn, entity)

    def _create_work_product_entities_nodes(self, work_product: dict):
        for part_name, work_product_part in work_product.items():
            if part_name == "WorkProduct":
                self._create_entity_node(work_product_part)
            else:
                for p in work_product_part:
                    self._create_entity_node(p)

    def _fill_srn_node_table(self):
        for entity in self.entities:
            self._create_entity_node(entity)

    def _fill_nodes_parents(self):
        """
        Find parents in every entity.
        """
        for entity_node in self.srn_node_table.values():
            self._set_entity_parents(entity_node)

    @tenacity.retry(
        wait=tenacity.wait_fixed(TIMEOUT),
        stop=tenacity.stop_after_attempt(RETRIES),
        reraise=True
    )
    @authorize()
    def _get_storage_record_request(self, headers: dict, srn: str) -> requests.Response:
        logger.debug(f"Searching for {srn}")
        return requests.get(f"{self.storage_service_url}/{srn}", headers=headers)

    def _is_in_storage(self, parent_srn: str) -> bool:
        try:
            self._get_storage_record_request(self.request_headers, parent_srn)
            return True
        except requests.HTTPError:
            return False

    def _set_entity_parents(self, entity: EntityNode):
        """
        Find all references parent in entity's content.
        If a parent is not presented in manifest, mark this entity as unprocessed.
        If a parent is not in manifest and it is already in Storage, we do nothing with it.
        """
        parent_srns = entity.get_parent_srns()
        for parent_srn in parent_srns:
            if self.srn_node_table.get(parent_srn):
                parent_node = self.srn_node_table[parent_srn]
                parent_node.add_child(entity)
                entity.add_parent(parent_node)
            elif self._is_in_storage(parent_srn):
                continue
            else:  # if entity refers to srn not presenting in manifest
                self.unprocessed_entities_parent.add_child(entity)
                logger.info(f"'{entity}' is orphaned. Missing parent '{parent_srn}'")

    def _find_unprocessed_nodes(self):
        """
        Traverse entities dependant on orphaned or invalid ones.
        Add them to set of unprocessed nodes to exclude them from ingestion queue.
        """
        queue = deque()
        queue.append(self.unprocessed_entities_parent)
        while queue:
            node = queue.popleft()
            self.unprocessed_entities.add(node)
            logger.debug(f"Node {node} added to unprocessed.")
            for child in node.children:
                if not child.unprocessed:
                    child.unprocessed = True
                    queue.append(child)
        self.unprocessed_entities.discard(self.unprocessed_entities_parent)

    def entity_queue(self) -> Iterator[EntityNode]:
        """
        Create a queue, where a child entity goes after all its parents.
        If an entity is marked as unprocessed, then skip it.
        """
        entity_graph = {entity: entity.parents for entity in self.srn_node_table.values()}
        logger.debug(f"Entity graph {entity_graph}.")
        entity_queue = toposort.toposort_flatten(entity_graph, sort=False)
        for entity in entity_queue:
            if entity not in self.unprocessed_entities:
                self.processed_entities.append(entity)
                yield entity

        for entity in self.unprocessed_entities:
            entity.replace_parents_surrogate_srns()

        logger.debug(f"Visited entities {self.processed_entities}")
        logger.debug(f"Unprocessed entities {self.unprocessed_entities}")

    def add_unprocessed_entity(self, entity: EntityNode):
        """
        Use if there some problems with ingesting entity.
        Mark it and its dependants as unprocessed.
        """
        self.unprocessed_entities_parent.add_child(entity)
        self._find_unprocessed_nodes()
